unit vb_au;

interface

uses classes, vb_pvflib;

const
     SND_MAGIC        = $2e736e64;
     SND_HEADER_SIZE  = 28;
     SND_UNKNOWN_SIZE = -1;

     BIAS = $84;    //* define the add-in bias for 16 bit samples */
     CLIP = 32635;

     sound_format: Array[0..8] of String = ('unspecified',
                                            'uLaw_8',
                                            'linear_8',
                                            'linear_16',
                                            'linear_24',
                                            'linear_32',
                                            'float',
                                            'double',
                                            'fragmented');
Type

  SNDSoundStruct = record
                                   // must all be 32bit
     magic: LongInt;               //* magic number SND_MAGIC */
     dataLocation: LongInt;        //* offset or pointer to the data */
     dataSize: LongInt;            //* number of bytes of data */
     dataFormat: LongInt;          //* the data format code */
     samplingRate: LongInt;        //* the sampling rate */
     channelCount: LongInt;        //* the number of channels */
     info: LongInt;                //* optional text information */
  End;

function  linear2ulaw(sample: integer): Byte;
function  ulaw2linear (ulawbyte: Byte): integer;
function  read_word(Stream: TMemoryStream): LongInt;
procedure write_word(data: LongInt; Stream: TMemoryStream);
function  pvftoau(Stream_in, Stream_out: TMemoryStream; header_in: Tpvf_header; dataFormat: integer): boolean;
function  autopvf(Stream_in, Stream_out: TMemoryStream; header_out: Tpvf_header): boolean;
function  pvftobasic(Stream_in, Stream_out: TMemoryStream; header_in: Tpvf_header): boolean;
function  basictopvf(Stream_in, Stream_out: TMemoryStream; header_out: Tpvf_header): boolean;

implementation

uses Dialogs, SysUtils;

(*
 * This routine converts from linear to ulaw.
 *
 * Craig Reese: IDA/Supercomputing Research Center
 * Joe Campbell: Department of Defense
 * 29 September 1989
 *
 * References:
 * 1) CCITT Recommendation G.711  (very difficult to follow)
 * 2) "A New Digital Technique for Implementation of Any
 *     Continuous PCM Companding Law," Villeret, Michel,
 *     et al. 1973 IEEE Int. Conf. on Communications, Vol 1,
 *     1973, pg. 11.12-11.17
 * 3) MIL-STD-188-113,"Interoperability and Performance Standards
 *     for Analog-to_Digital Conversion Techniques,"
 *     17 February 1987
 *
 * Input: Signed 16 bit linear sample
 * Output: 8 bit ulaw sample
 *)

function  linear2ulaw(sample: integer): Byte;
const
     exp_lut: Array[0..255] of integer = (
                      0, 0, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3,
                      4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
                      5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
                      5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
                      6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
                      6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6 ,6,
                      6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
                      6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
                      7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
                      7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
                      7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
                      7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
                      7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
                      7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
                      7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
                      7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7);

Var
   sign,
   exponent,
   mantissa: integer;
   ulawbyte: Byte;
Begin
//*
//* Get the sample into sign-magnitude.
//*

 sign := (sample shr 8) and $80;         //* set aside the sign */

 if (sign <> 0) then sample := -sample;  //* get magnitude */

 if (sample > CLIP) then sample := CLIP; //* clip the magnitude */

//*
//* Convert from 16 bit linear to ulaw.
//*

 sample   := sample + BIAS;
 exponent := exp_lut[(sample shr 7 ) and $ff];
 mantissa := (sample shr (exponent + 3)) and $0f;
 ulawbyte := not (sign or (exponent shl 4) or mantissa);

 (*
#ifdef ZEROTRAP

     if (ulawbyte == 0)
          ulawbyte = 0x02;          /* optional CCITT trap */

#endif
*)
 result := ulawbyte;
End;

(*
 * This routine converts from ulaw to 16 bit linear.
 *
 * Craig Reese: IDA/Supercomputing Research Center
 * 29 September 1989
 *
 * References:
 * 1) CCITT Recommendation G.711  (very difficult to follow)
 * 2) MIL-STD-188-113,"Interoperability and Performance Standards
 *     for Analog-to_Digital Conversion Techniques,"
 *     17 February 1987
 *
 * Input: 8 bit ulaw sample
 * Output: signed 16 bit linear sample
 *)
function  ulaw2linear (ulawbyte: Byte): integer;
const
     exp_lut: Array[0..7] of integer = (0, 132, 396, 924, 1980, 4092, 8316, 16764);
Var
   sign,
   exponent,
   mantissa,
   sample: integer;
Begin
 ulawbyte := (not ulawbyte);
 sign     := (ulawbyte and $80);
 exponent := (ulawbyte shr 4) and $07;
 mantissa := ulawbyte and $0f;
 sample   := exp_lut[exponent] + (mantissa shl (exponent + 3));

 if (sign <> 0) then sample := -sample;

 result := sample;
End;

function  read_word(Stream: TMemoryStream): LongInt;
Var
   data:   array[0..3] of Byte;
   MyWord: LongInt;
   MyByte: Byte;
   Count:  Integer;
Begin
 For Count := 0 To 3 Do
  Begin
   Stream.Read(MyByte,1);
   data[count] := MyByte;
  End;
 MyWord := data[0];
 MyWord := (MyWord shl 8) or data[1];
 MyWord := (MyWord shl 8) or data[2];
 MyWord := (MyWord shl 8) or data[3];
 result := MyWord;
End;

procedure write_word(data: LongInt; Stream: TMemoryStream);
Var
   outdata:   array[0..3] of Byte;
   Count:     Integer;
Begin
 outdata[0] := ((data and $ff000000) shr 24);
 outdata[1] := ((data and $00ff0000) shr 16);
 outdata[2] := ((data and $0000ff00) shr 8);
 outdata[3] := (data and $000000ff);
 For Count := 0 To 3 Do Stream.Write(outdata[Count],1);
End;

function  pvftoau(Stream_in, Stream_out: TMemoryStream; header_in: Tpvf_header; dataFormat: integer): boolean;
Var
   hdr: SNDSoundStruct;
   sample: integer;
   MyByte: Byte;
//   MySmall: SmallInt;
//   MyShort: ShortInt;
//   MyLong:  LongInt;
Begin
 hdr.magic        := SND_MAGIC;
 hdr.dataLocation := SND_HEADER_SIZE;
 hdr.dataSize     := SND_UNKNOWN_SIZE;
 hdr.dataFormat   := dataFormat;
 hdr.samplingRate := header_in.Speed;
 hdr.channelCount := 1;
 hdr.info         := 0;

 write_word(hdr.magic, Stream_out);
 write_word(hdr.dataLocation, Stream_out);
 write_word(hdr.dataSize, Stream_out);
 write_word(hdr.dataFormat, Stream_out);
 write_word(hdr.samplingRate, Stream_out);
 write_word(hdr.channelCount, Stream_out);
 write_word(hdr.info, Stream_out);

 Case hdr.dataFormat of
     SND_FORMAT_MULAW_8:   Begin
                            While Stream_in.Position < Stream_in.Size Do
                             Begin
                              sample := header_in.read_pvf_data(Stream_in);
                              sample := (sample shr 8);
                              if (sample < -32768) or (sample > 32767) then sample := SmallInt(sample);
                              MyByte := (linear2ulaw(sample) and $ff);
                              Stream_out.Write(MyByte,1);
                             End;
                           End;
     SND_FORMAT_LINEAR_8:  Begin
                            While Stream_in.Position < Stream_in.Size Do
                             Begin
                              sample := (header_in.read_pvf_data(Stream_in) shr 16);
                              if (sample < -32768) or (sample > 32767) then sample := SmallInt(sample);
                              MyByte := (sample and $ff);
                              Stream_out.Write(MyByte,1);
                             End;
                           End;
     SND_FORMAT_LINEAR_16: Begin
                            While Stream_in.Position < Stream_in.Size Do
                             Begin
                              sample := (header_in.read_pvf_data(Stream_in) shr 8);
                              if (sample < -128) or (sample > 127) then sample := SmallInt(sample);
                              MyByte := ((sample shr 8) and $ff);
                              Stream_out.Write(MyByte,1);
                              MyByte := (sample and $ff);
                              Stream_out.Write(MyByte,1);
                             End;
                           End;
     Else Begin
           Result := False;
           Exit;
          End;
 End;
 Result := True;
End;

function  autopvf(Stream_in, Stream_out: TMemoryStream; header_out: Tpvf_header): boolean;
Var
   hdr: SNDSoundStruct;
   sample: integer;
//   i:      integer;
   MyByte: Byte;
//   MyChar: Char;
Begin
 hdr.magic        := read_word(Stream_in);
 hdr.dataLocation := read_word(Stream_in);
 hdr.dataSize     := read_word(Stream_in);
 hdr.dataFormat   := read_word(Stream_in);
 hdr.samplingRate := read_word(Stream_in);
 hdr.channelCount := read_word(Stream_in);
 //* hdr.info = read_word(fd_in); */ /* this is sometimes missing */

 if (hdr.magic <> SND_MAGIC) then
  Begin
   result := false;
   Exit
  End;

 if (hdr.channelCount <> 1) then
  Begin
   result := false;
   Exit
  End;

 header_out.Speed := hdr.samplingRate;

 if not header_out.write_pvf_header(Stream_out) Then
  Begin
   result := false;
   Exit
  End;

 If hdr.dataLocation >= Stream_in.Size Then
  Begin
   result := false;
   Exit
  End
 Else Stream_in.Position := hdr.dataLocation;

 case hdr.dataFormat of
     SND_FORMAT_MULAW_8:   Begin
                            While Stream_in.Position < Stream_in.Size Do
                             Begin
                              Stream_in.Read(MyByte,1);
                              sample := (ulaw2linear(MyByte) shl 8);
                              header_out.write_pvf_data(Stream_out,sample);
                             End;
                           End;
     SND_FORMAT_LINEAR_8:  Begin
                            While Stream_in.Position < Stream_in.Size Do
                             Begin
                              Stream_in.Read(MyByte,1);
                              sample := (MyByte and $ff);
                              If (sample > $7f) then sample := sample - $100;
                              header_out.write_pvf_data(Stream_out,sample shl 16);

//                              Stream_in.Read(MyByte,1);
//                              if (MyByte > $7f) then sample := MyByte - $100
//                                                Else sample := MyByte;
//                              sample := SmallInt(sample shl 16);
//                              sample := sample and $ff;
//                              header_out.write_pvf_data(Stream_out,sample);
                             End;
                           End;
     SND_FORMAT_LINEAR_16: Begin
                            While Stream_in.Position < Stream_in.Size Do
                             Begin
                              Stream_in.Read(MyByte,1);
                              sample := MyByte;
//                              sample := sample and $ffff;
                              if (sample > $7fff) then sample := sample - $10000;
                              header_out.write_pvf_data(Stream_out,sample shl 8);
                             End;
                           End;
     Else Begin
           Result := False;
           Exit;
          End;

 End;
 Result := True;
End;

function  pvftobasic(Stream_in, Stream_out: TMemoryStream; header_in: Tpvf_header): boolean;
Var
   sample: integer;
Begin
 if (header_in.Speed <> 8000) Then
  Begin
   Result := False;
   Exit;
  End;
 While Stream_in.Position < Stream_in.Size Do
  Begin
   sample := (header_in.read_pvf_data(Stream_in) shr 8);
   sample := linear2ulaw(sample);
   Stream_out.Write(sample,1);
  End;
 Result := True;
End;

function  basictopvf(Stream_in, Stream_out: TMemoryStream; header_out: Tpvf_header): boolean;
Var
   sample: integer;
   MyChar: Char;
Begin
 if (header_out.Speed <> 8000) Then
  Begin
   Result := False;
   Exit;
  End;
 While Stream_in.Position < Stream_in.Size Do
  Begin
   Stream_in.Read(MyChar,1);
   sample := ord(MyChar);
   sample := (ulaw2linear(sample) shl 8);
   header_out.write_pvf_data(Stream_out, sample);
  End;
 Result := True;
End;

end.
